"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createRemixRequest = exports.createRemixHeaders = exports.createRequestHandler = void 0;
const node_1 = require("@remix-run/node");
/**
 * Returns a request handler for Google Cloud functions that serves the response using Remix.
 */
function createRequestHandler({ build, getLoadContext, mode = process.env.NODE_ENV, }) {
    let handleRequest = (0, node_1.createRequestHandler)(build, mode);
    return async (req, res) => {
        try {
            let request = createRemixRequest(req, res);
            let loadContext = typeof getLoadContext === "function"
                ? getLoadContext(req, res)
                : undefined;
            let response = (await handleRequest(request, loadContext));
            await sendRemixResponse(res, response);
        }
        catch (error) {
            console.error(error);
            await sendRemixResponse(res, new node_1.Response("Internal Error", { status: 500 }));
        }
    };
}
exports.createRequestHandler = createRequestHandler;
function createRemixHeaders(requestHeaders) {
    let headers = new node_1.Headers();
    for (let [key, values] of Object.entries(requestHeaders)) {
        if (values) {
            if (Array.isArray(values)) {
                for (let value of values) {
                    headers.append(key, value);
                }
            }
            else {
                headers.set(key, values);
            }
        }
    }
    return headers;
}
exports.createRemixHeaders = createRemixHeaders;
function createRemixRequest(req, res) {
    let origin = `${req.protocol}://${req.get("host")}`;
    let url = new URL(req.url, origin);
    let controller = new node_1.AbortController();
    res.on("close", () => controller.abort());
    let init = {
        method: req.method,
        headers: createRemixHeaders(req.headers),
        signal: controller.signal,
    };
    if (req.method !== "GET" && req.method !== "HEAD") {
        init.body = req.rawBody;
    }
    return new node_1.Request(url.href, init);
}
exports.createRemixRequest = createRemixRequest;
async function sendRemixResponse(res, nodeResponse) {
    res.statusMessage = nodeResponse.statusText;
    res.status(nodeResponse.status);
    for (let [key, values] of Object.entries(nodeResponse.headers.raw())) {
        for (let value of values) {
            res.append(key, value);
        }
    }
    if (nodeResponse.body) {
        await (0, node_1.writeReadableStreamToWritable)(nodeResponse.body, res);
    }
    else {
        res.end();
    }
}
